
import os
import toml
import json

from rich.prompt import Confirm

from ..utils import get_vcpkg_root, pkg_exists, copy_directory
from ..config import Package, read_package_config
from ..paths import *



# Adds the package to the project directory
# @param package_name (str): Name of the package
def add_pkg(package_name: str) -> None:

    if not os.path.exists(PROJ_CONFIG_FILE):
        print("Error: Project configuration file not found")
        return 

    if pkg_exists(package_name):
        print("Error: Package already added!")
        return
    
    pkg_dir_path = get_installed_pkg_dir_path(package_name)
    
    if not pkg_dir_path:
        if Confirm.ask("Would you like me to install this package for you?"):
            os.system(f'vcpkg install {package_name}')
            add_pkg(package_name)
        return
    
    if not is_pkg_valid(pkg_dir_path):
        print("Error: Not a valid package")
        return
    
    copy_pkg(package_name, pkg_dir_path) 
    add_pkg_to_requirements(read_package_config(os.path.join(DIRS["LIB_PKGCONFIG_DIR"], package_name + '.pc')))

    print('Added:', package_name, 'package')


# Copies all the pacakge files like include and lib to the project directory path
# @param package_name (str): name of the package
# @param package_path (str): path of the package
def copy_pkg(package_name: str, package_path: str) -> None:

    dirs_to_copy = {
        "LIB_DIR": os.path.join(package_path, "lib"),
        "INCLUDE_DIR": os.path.join(package_path, "include"),
        "DEBUG_LIB_DIR": os.path.join(package_path, "debug/lib"),
    }

    copied_contents: list[str] = []

    for dest, src in dirs_to_copy.items():
        try:
            copied_contents.append(copy_directory(src, os.path.join(DIRS[dest])))
            if not copied_contents[-1]:
                raise Exception("Error while copying files")
        except Exception as e:
            print(f"Error (while adding package): {e}")
            print("Reverting changes, Cleaning up...")
            for copied_dir in copied_contents:
                for file in copied_dir:
                    os.remove(file)
            return
        
    pkg_index_data: list[dict] = []

    if os.path.exists(PKG_INDEX_FILE):
        with open(PKG_INDEX_FILE, 'r') as f:
            pkg_index_data = json.load(f)

    pkg_index_data.append({package_name: copied_contents})
    with open(PKG_INDEX_FILE, 'w') as f:
        json.dump(pkg_index_data, f, indent=2)


# Scans vcpkg root directory and finds the package directory for given the package name
# @param package_name (str): name of the package
# @Returns: (str) package path if package is installed | None if not installed
def get_installed_pkg_dir_path(package_name: str) -> str|None:

    with os.scandir(os.path.join(get_vcpkg_root(), "packages")) as entries:
        for entry in entries:
            if entry.is_dir():
                pkg_name = entry.name.split("_")[0]
                if pkg_name == package_name:
                    return entry.path
                
    return None
            

# Checks if the package contains the necessery files and directories
# @returns: True if pkg is valid | False if pkg not valid
def is_pkg_valid(package_path: str) -> bool:
    is_valid: bool = True

    if not os.path.exists(os.path.join(package_path, "include")):
        is_valid = False
    if not os.path.exists(os.path.join(package_path, "debug")):
        is_valid = False
    if not os.path.exists(os.path.join(package_path, "lib")):
        is_valid = False

    return is_valid


# Adds the package to the requirements table in the project config
# @param: Package object
def add_pkg_to_requirements(package: Package):

    with open(PROJ_CONFIG_FILE, "r") as f:
        data = toml.load(f)

    data.setdefault("requirements", {}).setdefault("package", []).append({
        "name": package.name,
        "description": package.description,
        "url": package.url,
        "version": package.version
    })

    with open(PROJ_CONFIG_FILE, "w") as f:
        toml.dump(data, f)



import os
import json
import pathlib
import subprocess


from .paths import *



# @return: the root directory of vcpkg
def get_vcpkg_root() -> str|Exception:
    vcpkg_root = os.environ.get('VCPKG_ROOT')
    if vcpkg_root is None:
        raise Exception('VCPKG_ROOT is not set')
    return vcpkg_root


# @returns: str cc or c++ based on the file extention
def get_compiler_type(file: str|list[str]) -> str:
    if isinstance(file, str):
        if pathlib.Path(file).suffix == '.c':
            return'cc'
        elif pathlib.Path(file).suffix == '.cpp':
            return 'c++'
    elif isinstance(file, list):
        has_cpp = (pathlib.Path(f).suffix == '.cpp' for f in file)
        return 'c++' if has_cpp else 'cc'


# Runs the zig compile command and compiles the file
# @param src_file (str): filepath of the source file
def zig_compile(src_file: str, cflags: list[str], additional_flags: str = "") -> bool:
    print('[ compilation started ]', pathlib.Path(src_file).name)
    cmd = ['zig', get_compiler_type(src_file), '-c', additional_flags, *cflags, src_file, '-o', os.path.join(DIRS["CACHE_DIR"], src_file.encode('utf-8').hex() + '.o')]
    result = subprocess.run(cmd, capture_output=True, text=True)
    
    if result.returncode != 0:
        print('[ compilation error ]', pathlib.Path(src_file).name, '->', result.stderr)
        return False
    print('[ compilation finished ]', pathlib.Path(src_file).name)
    return True


# Runs the zig command to create an executable
# @param input_files list[str]: 
# @param target str: target of platfrom for creating the executable
def zig_link(input_files: list[str], output_path: str, libs: list[str], cflags: list[str] = [], target: str = 'native', additional_flags: str = "") -> None:
    print(f'Linking: {len(input_files)} objects | target={target}')
    cmd = ['zig', get_compiler_type(input_files), '-target', target, additional_flags, *input_files, *cflags, *libs, '-o', output_path]
    result = subprocess.run(" ".join(cmd), shell=True, capture_output=True, text=True)
    if result.returncode != 0:
        print(result.stderr)


# Checks if package exists in the pkg-index.json file.
# If it does then it checks if all the files in the pkg exists or not
# @param package_name (str): name of the package.
# @return: True if exists else False
def pkg_exists(package_name: str) -> bool:

    pkg_index_data: list[dict] = []

    if os.path.exists(PKG_INDEX_FILE):
        with open(PKG_INDEX_FILE, 'r') as f:
            pkg_index_data = json.load(f)

    for pkg in pkg_index_data:
        for pkg_name in pkg.keys():
            if pkg_name == package_name:
                for pkg_dirs in pkg.values():
                    for dir in pkg_dirs:
                        for file in dir:
                            if not os.path.exists(file):
                                return False
                    return True
            break

    return False

# Copies a directory from the source path to the destination path.
# @param src_dir (str): The source directory path.
# @param dest_dir (str): The destination directory path.
# @return: A list of the directories that were copied.
def copy_directory(src_dir, dest_dir) -> list[str]:
    import shutil

    copied_so_far: list[str] = []

    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)

    for root, _, files in os.walk(src_dir):
        rel_path = os.path.relpath(root, src_dir)

        dest_subdir = os.path.join(dest_dir, rel_path)
        if not os.path.exists(dest_subdir):
            os.makedirs(dest_subdir)
            copied_so_far.append(dest_subdir)

        for file in files:
            src_file = os.path.join(root, file)
            dest_file = os.path.join(dest_subdir, file)
            shutil.copy2(src_file, dest_file)
            copied_so_far.append(dest_file)

    return copied_so_far




# Reads the debug/bin/cache directory and creates a list of .o (object) files
# Returns: The list of object file paths
def get_object_files() -> dict:

    o_files: dict = {}

    for file in pathlib.Path(DIRS["DEBUG_BIN_CACHE_DIR"]).iterdir():
        if file.is_file():
            if file.suffix == '.o':
                o_files[file.name] = str(file.absolute())

    return o_files


# Reads the /src directory and creates a list of .c and .cpp files
# Returns: The list of source file paths
def get_src_files() -> list[str]:

    src_files: list[str] = []

    for file in pathlib.Path(DIRS["SRC_DIR"]).rglob("*.c"):
        src_files.append(str(file.absolute()))

    for file in pathlib.Path(DIRS["SRC_DIR"]).rglob("*.cpp"):
        src_files.append(str(file.absolute()))

    return src_files



# Checks if the project directory contains the project config file and zig is installed
# Returns: True if project is valid and zig is installed else False
def check_valid_proj_and_zig_installed() -> bool:
    
    if not os.path.exists(PROJ_CONFIG_FILE):
        print("Error: Project config file not found")    
        return False

    try:
        ver = subprocess.run(["zig", "version"], capture_output=True, text=True)
    except FileNotFoundError as e:
        print("Error: zig is not installed")
        return False
    else:
        return True


import os

cwd = os.getcwd()                          # Current working directory

DIRS = {
    "CANDIE_DIR": f"{cwd}/.candie",                          # Candie directory
    "CACHE_DIR": f"{cwd}/.candie/cache",                     # cache directory

    "SRC_DIR": f"{cwd}/src",                                 # Source code directory
    "BUILD_DIR": f"{cwd}/build",                             # Build directory
    "INCLUDE_DIR": f"{cwd}/include",                         # Include directory

    "LIB_DIR": f"{cwd}/lib",                                 # Library directory
    "LIB_PKGCONFIG_DIR": f"{cwd}/lib/pkgconfig",             # Library pkgconfig directory

    "DEBUG_DIR": f"{cwd}/debug",                             # Debug build directory

    "DEBUG_LIB_DIR": f"{cwd}/debug/lib",                     # Debug library directory
    "DEBUG_LIB_PKGCONFIG_DIR": f"{cwd}/debug/lib/pkgconfig", # Debug library pkgconfig directory

    "DEBUG_BIN_DIR": f"{cwd}/debug/bin",                     # Debug binary directory
}

MODIF_LOG_FILE = f"{cwd}/.candie/log-file.json"               # Log file path
PROJ_CONFIG_FILE = f"{cwd}/proj-config.toml"                  # Project configuration file path

PKG_INDEX_FILE = f"{cwd}/.candie/pkg-index.json"              # Package index file path


import typer


app = typer.Typer()


@app.command(help="Create a C/C++ project")
def create(project_name: str):
    from .commands.create import create_proj
    create_proj(project_name)


@app.command(help="Add a package to the project")
def add(package_name: str):
    from .commands.add import add_pkg
    add_pkg(package_name)


@app.command(help="Remove a package from the project")
def remove(package_name: str):
    from .commands.remove import remove_pkg
    remove_pkg(package_name)


@app.command(help="Make the debug application")
def make():
    from .commands.make import make_app
    make_app()


@app.command(help="Run the debug application")
def run(
    remake: bool = typer.Option(False, "--remake", help="Recompile and run the program"),
):
    from .commands.run import run_prog
    run_prog(remake)


@app.command(help="Build the project")
def build():
    from .commands.build import build_proj
    build_proj()


@app.command(help="Print the version")
def version():
    print('version: 1.1.5')


def candie_exec():
    app()


import toml
import platform


from .paths import *


class Package:
    def __init__(self):
        self.name: str         = ''      # name of the package
        self.description: str  = ''      # package decription
        self.url: str          = ''      # package url
        self.version: str      = ''      # package version

        self.libs: str         = ''      # package libraries linking flags
        self.cflags: str       = ''      # compilation flags



# Creates project config file
# writes necessary values to the config file after creating it.
# @param name (str): name of the project
# @param description (str): project description
def create_proj_config_file(name: str, description: str = "") -> None:

    config: dict = {}

    config["project"] = {
        "name": name,
        "description": description,
        "version": "0.0.1",
    }

    config["debug"] = {
        'Cflags': "-Wall -Wextra -g",
        'Lflags': '-Wl,--no-undefined'
    }

    config["build"] = {
        'flags': '-Wl,--no-undefined',
        'target': [
            {
                'arch': platform.machine(),
                'os': platform.system().lower()
            }
        ]
    }

    with open(PROJ_CONFIG_FILE, 'w') as f:
        toml.dump(config, f)



# @returns: project config
def get_proj_config() -> dict:
    with open(PROJ_CONFIG_FILE, 'r') as f:
        return toml.load(f)


# Reads the <package>.pc file in the pkgconfig directory and returns the package object.
# @param pc_file_path (str): filepath of the .pc file
# @returns: Package or none if error occurs
def read_package_config(pc_file_path: str) -> Package|None:
    try:
        with open(pc_file_path, 'r') as file:
            package = Package()
            for line in file:
                line = line.strip()
                if not line or ":" not in line:
                    continue
                key, value = line.split(":", 1)
                key = key.strip()
                value = value.strip()
                if hasattr(package, key.lower()):
                    setattr(package, key.lower(), value)
            if not package.name:
                print("Error (not a package): package name not found in config file")
                return None
            
            package.name = os.path.basename(pc_file_path).removesuffix('.pc')

            return package
    except Exception as e:
        print(f"Error reading pc file: {e}")
        return None



import os
import subprocess

from rich import print as rprint
 
from .make import make_app
from ..paths import *
from ..config import get_proj_config
from ..utils import check_valid_proj_and_zig_installed


# Compiles the project and Runs the executable.
def run_prog(remake: bool) -> None:

    # validate project and requirements
    if not check_valid_proj_and_zig_installed():
        return

    # get project config
    proj_config = get_proj_config()

    # Create program path
    program = os.path.join(DIRS["DEBUG_BIN_OUTPUT_DIR"], proj_config['project']['name'])

    # Recompile if remake true
    if remake:
        if not make_app():
            return

    # If program exists then run
    if os.path.exists(program):
        subprocess.run([program])
    else: 
        rprint("Error: Executable not found")


import os

from rich.prompt import Prompt


from ..paths import *
from ..config import create_proj_config_file

# Creates a new project 
# with the given name
# @param project_name: str
def create_proj(project_name: str = '') -> None:

    if os.listdir(cwd):
        print("Error: Directory not empty!")
        return
    
    if project_name == '':
        project_name = Prompt.ask("Please enter Project Name")

    with open('.gitignore', 'w') as f:
            f.write('''
.candie/
build/
debug/
include/
lib/
''')
    
    for dir in DIRS.values():
        os.makedirs(dir, exist_ok=True)
    
    with open(os.path.join(DIRS["SRC_DIR"], 'main.c'), 'w') as f:
                f.write('''
#include<stdio.h>

int main() {
    printf("Hello, World!");
    return 0;
}
''')
    
    create_proj_config_file(project_name)

    with open(MODIF_LOG_FILE, 'w') as f:
        f.write('{}')

    print('Project created.')
    print('Run [candie make] to compile.\nExecute the program using [candie run] command')
